#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdint.h>
#include <string.h>
#include <fcntl.h>
#include <time.h>
#include <getopt.h>
#include <errno.h>
#include <sys/wait.h>
#include <scsi/scsi.h>
#include <scsi/scsi_ioctl.h>
#include <scsi/sg.h>
#include <sys/ioctl.h>
#include "diag_encl.h"

static struct option long_options[] = {
	{"help",		no_argument,		NULL, 'h'},
	{"serv_event",		no_argument,		NULL, 's'},
	{"verbose",		no_argument,		NULL, 'v'},
	{0,0,0,0}
};

static struct {
	char *mtm;
	int (*func)(int, struct dev_vpd *, int, int);
} encl_diags[] = {
	{"7031-D24/T24", diag_7031_D24_T24},	/* Pearl enclosure */
	{NULL, NULL},
	};


/**
 * print_usage
 * @brief Print the usage message for this command
 *
 * @param name the name of this executable
 */ 
static void
print_usage(const char *name) {
	printf("Usage: %s [-h] [-s] [-v] [<scsi_enclosure>]\n"
		"\t-h: print this help message\n"
		"\t-s: generate a serviceable event for any failures and\n"
		"\t      write events to the servicelog\n"
		"\t-v: verbose output\n"
		"\t<scsi_enclosure>: the sg device on which to operate, such\n"
		"\t      as sg7; if not specified, all such devices will be\n"
		"\t      diagnosed\n", name);
	return;
}

/**
 * print_raw_data
 * @brief Dump a section of raw data 
 *
 * @param ostream stream to which output should be written
 * @param data pointer to data to dump
 * @param data_len length of data to dump
 * @return number of bytes written
 */ 
int
print_raw_data(FILE *ostream, char *data, int data_len)
{
	char *h, *a;
	char *end = data + data_len;
	unsigned int offset = 0;
	int i, j;
	int len = 0;

	len += fprintf(ostream, "\n");

	h = a = data;

	while (h < end) {
		/* print offset */
		len += fprintf(ostream, "0x%04x:  ", offset);
		offset += 16;

		/* print hex */
		for (i = 0; i < 4; i++) {
			for (j = 0; j < 4; j++) {
				if (h < end)
					len += fprintf(ostream, "%02x", *h++);
				else
					len += fprintf(ostream, "  ");
			}
			len += fprintf(ostream, " ");
		}

		/* print ascii */
		len += fprintf(ostream, "    [");
		for (i = 0; i < 16; i++) {
			if (a <= end) {
				if ((*a >= ' ') && (*a <= '~'))
					len += fprintf(ostream, "%c", *a);
				else
					len += fprintf(ostream, ".");
				a++;
			} else
				len += fprintf(ostream, " ");
		}
		len += fprintf(ostream, "]\n");
	}

	return len;
}

/**
 * add_callout
 * @brief Create a new sl_callout struct
 *
 * @param callouts address of pointer to callout list
 * @param priority callout priority
 * @param type callout type
 * @param proc_id callout procedure ID
 * @param location callout location code
 * @param fru callout FRU number
 * @param sn callout FRU serial number
 * @param ccin callout FRU ccin
 */
void
add_callout(struct sl_callout **callouts, char priority, uint32_t type,
	    char *proc_id, char *location, char *fru, char *sn, char *ccin)
{
	struct sl_callout *c;

	if (*callouts == NULL) {
		*callouts = malloc(sizeof(struct sl_callout));
		c = *callouts;
	}
	else {
		c = *callouts;
		while (c->next != NULL)
			c = c->next;
		c->next = malloc(sizeof(struct sl_callout));
		c = c->next;
	}
	if (c == NULL)
		return;

	memset(c, 0, sizeof(struct sl_callout));
	c->priority = priority;
	c->type = type;
	if (proc_id) {
		c->procedure = (char *)malloc(strlen(proc_id) + 1);
		strcpy(c->procedure, proc_id);
	}
	if (location) {
		c->location = (char *)malloc(strlen(location) + 1);
		strcpy(c->location, location);
	}
	if (fru) {
		c->fru = (char *)malloc(strlen(fru) + 1);
		strcpy(c->fru, fru);
	}
	if (sn) {
		c->serial = (char *)malloc(strlen(sn) + 1);
		strcpy(c->serial, sn);
	}
	if (ccin) {
		c->ccin = (char *)malloc(strlen(ccin) + 1);
		strcpy(c->ccin, ccin);
	}
}

/**
 * servevent
 * @brief Generate a serviceable event and an entry to the servicelog
 *
 * @param refcode the SRN or SRC for the serviceable event
 * @param sev the severity of the event
 * @param text the description of the serviceable event
 * @param vpd a structure containing the VPD of the target
 * @param callouts a linked list of FRU callouts
 * @return key of the new servicelog entry
 */ 
uint32_t
servevent(char *refcode, int sev, char *text, struct dev_vpd *vpd,
	  struct sl_callout *callouts)
{
	struct servicelog *slog;
	struct sl_event *entry;
	struct sl_data_enclosure *encl;
	uint64_t key;
	int rc;

	if ((refcode == NULL) || (text == NULL) || (vpd == NULL))
		return 0;

	entry = (struct sl_event *)malloc(sizeof(struct sl_event));
	memset(entry, 0, sizeof(struct sl_event));

	encl = (struct sl_data_enclosure *)malloc(
					sizeof(struct sl_data_enclosure));
	memset(encl, 0, sizeof(struct sl_data_enclosure));

	entry->addl_data = encl;

	entry->time_event = time(NULL);
	entry->type = SL_TYPE_ENCLOSURE;
	entry->severity = sev;
	entry->serviceable = 1;
	entry->call_home_status = 1;
	entry->description = (char *)malloc(strlen(text) + 1);
	strcpy(entry->description, text);
	entry->refcode = (char *)malloc(strlen(refcode) + 1);
	strcpy(entry->refcode, refcode);
	encl->enclosure_model = (char *)malloc(strlen(vpd->mtm) + 1);
	strcpy(encl->enclosure_model, vpd->mtm);
	encl->enclosure_serial = (char *)malloc(strlen(vpd->sn) + 1);
	strcpy(encl->enclosure_serial, vpd->sn);

	entry->callouts = callouts;

	rc = servicelog_open(&slog, 0);
	if (rc != 0) {
		fprintf(stderr, servicelog_error(slog));
		return 0;
	}
	rc = servicelog_event_log(slog, entry, &key);
	servicelog_event_free(entry);

	if (rc != 0)
		fprintf(stderr, servicelog_error(slog));

	servicelog_close(slog);

	return key;
}

/**
 * get_diagnostic_page
 * @brief Make the necessary sg ioctl() to retrieve a diagnostic page
 *
 * @param fd a file descriptor to the appropriate /dev/sg<x> file
 * @param diag_page the page to be retrieved
 * @param buf a buffer to contain the contents of the diagnostic page
 * @param buf_len the length of the previous parameter
 * @return 0 on success, -EIO on invalid I/O status, or CHECK_CONDITION
 */ 
int
get_diagnostic_page(int fd, int diag_page, void *buf, int buf_len)
{
	unsigned char scsi_cmd_buf[16] =
		{	RECEIVE_DIAGNOSTIC,
			0x01,			/* set PCV bit to 1 */
			diag_page,		/* page to be retrieved */
			(buf_len >> 8) & 0xff,	/* most significant byte */
			buf_len & 0xff,		/* least significant byte */
			0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
	struct sense_data_t sense_data;
	sg_io_hdr_t hdr;
	int i, rc;

	for (i = 0; i < 3; i++) {
		memset(&hdr, 0, sizeof(hdr));
		memset(&sense_data, 0, sizeof(struct sense_data_t));

		hdr.interface_id = 'S';
		hdr.dxfer_direction = SG_DXFER_FROM_DEV;
		hdr.cmd_len = 16;
		hdr.mx_sb_len = sizeof(sense_data);
		hdr.iovec_count = 0;	/* scatter gather not necessary */
		hdr.dxfer_len = buf_len;
		hdr.dxferp = buf;
		hdr.cmdp = scsi_cmd_buf;
		hdr.sbp = (unsigned char *)&sense_data;
		hdr.timeout = 120 * 1000;	/* set timeout to 2 minutes */
		hdr.flags = 0;
		hdr.pack_id = 0;
		hdr.usr_ptr = 0;

		rc = ioctl(fd, SG_IO, &hdr);

		if ((rc == 0) && (hdr.masked_status == CHECK_CONDITION))
			rc = CHECK_CONDITION;
		else if ((rc == 0) && (hdr.host_status || hdr.driver_status))
			rc = -EIO;

		if (rc == 0 || hdr.host_status == 1)
			break;
	}

	return rc;
}

/**
 * diagnose
 * @brief Diagnose a specific SCSI generic enclosure
 *
 * @param sg the SCSI generic device to diagnose (e.g. "sg7")
 * @param flag_s set to non-zero to generate a serviceable event for failures
 * @param flag_v set to non-zero to print verbose output
 * @return 0 for no failure, 1 if there is a failure on the enclosure
 */ 
static int
diagnose(const char *sg, int flag_s, int flag_v) {
	int rc = 0, fd, found = 0, i;
	char buf[128], devsg[128], *pos, *pos2;
	FILE *fp;
	struct dev_vpd vpd;

	printf("DIAGNOSING %s\n", sg);

	memset(&vpd, 0, sizeof(struct dev_vpd));

	/* use lscfg to find the MTM and location for the specified device */
	snprintf(buf, 128, "lscfg -vl %s", sg);
	if ((fp = popen(buf, "r")) == NULL) {
		printf("Unable to retrieve the MTM or location code.  "
				"Ensure that lsvpd is installed.\n\n");
		return rc;
	}
	while (fgets(buf, 128, fp) != NULL) {
		if ((pos = strstr(buf, "Machine Type and Model")) != NULL) {
			if ((pos2 = strchr(pos, '\n')) != NULL)
				*pos2 = '\0';
			pos += 28;
			strncpy(vpd.mtm, pos, 128);
		}
		else if ((pos = strstr(buf, "Device Specific.(YL)")) != NULL) {
			if ((pos2 = strchr(pos, '\n')) != NULL)
				*pos2 = '\0';
			pos += 28;
			strncpy(vpd.location, pos, 128);
		}
		else if ((pos = strstr(buf, "Serial Number")) != NULL) {
			if ((pos2 = strchr(pos, '\n')) != NULL)
				*pos2 = '\0';
			pos += 28;
			strncpy(vpd.sn, pos, 128);
		}
		else if ((pos = strstr(buf, "FN ")) != NULL) {
			if ((pos2 = strchr(pos, '\n')) != NULL)
				*pos2 = '\0';
			pos += 3;
			strncpy(vpd.fru, pos, 128);
		}
	}
	pclose(fp);

	printf("\t%s\n\t%s\n\n", vpd.mtm, vpd.location);

	snprintf(devsg, 128, "/dev/%s", sg);
	fd = open(devsg, O_RDWR);
	if (fd <= 1) {
		printf("Unable to open %s\n\n", devsg);
		return rc;
	}

	for (i=0; encl_diags[i].mtm; i++) {
		if (!strcmp(vpd.mtm, encl_diags[i].mtm)) {
			found = 1;
			rc += (*(encl_diags[i].func))(fd, &vpd, flag_s, flag_v);
		}
	}

	close(fd);

	if (!found)
		printf("Unable to diagnose devices of machine type/model "
			"%s\n\n", vpd.mtm);

	return rc;
}

int
main(int argc, char *argv[])
{
	int failure = 0, option_index, rc;
	int flag_s = 0, flag_v = 0;
	char buf[128], *sg, *pos;
	FILE *fp;

	for (;;) {
		option_index = 0;
		rc = getopt_long(argc, argv, "hsv", long_options,
				&option_index);

		if (rc == -1)
			break;

		switch (rc) {
		case 'h':
			print_usage(argv[0]);
			return 0;
		case 's':
			flag_s = 1;
			break;
		case 'v':
			flag_v = 1;
			break;
		case '?':
			print_usage(argv[0]);
			return -1;
			break;
		default:
			printf("huh?\n");
			break;
		}
	}

	if (optind < argc) {
		while (optind < argc) {
			failure += diagnose(argv[optind++], flag_s, flag_v);
		}
	}
	else {
		/* use lsvpd to find all sg devices */
		if ((fp = popen("lsvpd | grep /dev/sg", "r"))
				== NULL) {
			printf("Could not obtain a list of sg devices.  "
					"Ensure that lsvpd is installed.\n");
			return -1;
		}
		while (fgets(buf, 128, fp) != NULL) {
			if ((sg = strstr(buf, "sg")) != NULL) {
				if ((pos = strchr(sg, '\n')) != NULL)
					*pos = '\0';
				failure += diagnose(sg, flag_s, flag_v);
			}
		}
		pclose(fp);
	}

	return failure;
}
